"""Chart/visualization controllers.

Controllers for chart and visualization operators.
"""

# Import chart controllers for registration
from nirs4all.controllers.charts.exclusion import ExclusionChartController

__all__ = [
    "ExclusionChartController",
]
