presets = {
    'MinMaxScaler': ('sklearn.preprocessing', 'MinMaxScaler'),
    'StandardScaler': ('sklearn.preprocessing', 'StandardScaler'),
    'PCA': ('sklearn.decomposition', 'PCA'),
    'KMeans': ('sklearn.cluster', 'KMeans'),
    # Add more presets as needed
}
